"""
This module contains the base code and interface of basic TODS.
"""
from botiverse.bots.basic_TODS.utils import RuleBasedNLU, MostRecentDST, RandomDP, TemplateBasedNLG


class BasicTODS:
    """
    Instantiate a Task Oriented Dialogue System chat bot.
    It aims to assist the user in completing certain tasks in specific domains.
    The chat bot can use a fully Rule-Based approach which required no data for
    training and can also include Deep learning and data can be then used to train
    the chatbot model.

    :param name: The chatbot's name.
    :type name: str

    :param domains_slots: The slots of each domain in the system.
    :type domains_slots: dict[str, list[str]]

    :param templates: The predefined templates, where templates["domain"]["slot"] is
        a list of questions asking about slot "slot" in domain "domain".
    :type templates: dict[str, dict[str, list[str]]]

    :param domains_pattern: Dictionary where the key is a name of a domain
        and the value is a Regex that is used to capture that domain.
    :type domains_pattern: dict[str, str]

    :param slots_pattern: Dictionary where the key is a name of a domain
        and the value is another dictionary where the key is a name of a slot
        inside the domain and the value is a Regex that is used to capture that slot.
    :type slots_pattern: dict[str, dict[str, str]]

    :param is_classical: Indicates whether to use a fully classical approach, defaults to False
    :type is_classial: bool
    """

    def __init__(self, name, domains_slots, templates,
    domains_pattern, slots_pattern, is_classical=False):
        self.name = name
        self.nlu = RuleBasedNLU(domains_pattern, slots_pattern) if is_classical is False else None
        self.dst = MostRecentDST(domains_slots)
        self.dpo = RandomDP()
        self.nlg = TemplateBasedNLG(templates)
        self.current_domain = None

    def train(self, data):
        """
        Train the chatbot model with the given JSON data.

        :param data: A stringfied JSON object containing the training data
        :type number: string

        :return: None
        :rtype: NoneType
        """
        pass
    
    def infer(self, prompt):
        """
        Infer a suitable response to the given prompt.

        :param promp: The user's prompt
        :type number: str

        :return: When all slots in the current domain are filled it returns True, &
            the domain name, otherwise it returns False, & the chatbot's response.
        :rtype: tuple[bool, str]
        """

        status, self.current_domain = self.nlu.get_domain(self.current_domain, prompt)
        if status is False or self.current_domain is None:
            return False, "Sorry I don't know what domain are you talking about!"

        status, slots, values = self.nlu.get_slot_fillers(self.current_domain, prompt)
        if status is False:
            return False, "Sorry I Couldn't catch that can you say it in a different way!"

        self.dst.update_state(self.current_domain, slots, values)

        action = self.dpo.get_action(self.current_domain, self.dst.get_dialogue_state())
        if action == "":
            return True, self.current_domain

        response = self.nlg.generate(self.current_domain, action)

        return False, response

    def get_dialogue_state(self):
        """
        Gets the dialogue state.

        :return: The dialogue state, where state["domain"]["slot"] indicates
            value of slot "slot" in domain "domain".
        :rtype: dict[str, dict[str, str]]
        """
        return self.dst.get_dialogue_state()

    def reset(self, domain=None, slot=None):
        """
        Reset the dialogue state.
        Note: if the domain or the slot is equal to None all domains
        or slots will be reset respectively.

        :param domain: The domain to be reset, defaults to None.
        :type domain: str

        :param slot: The slot to be reset, defaults to None.
        :type slot: str

        """
        self.dst.reset(domain, slot)



"""
Example on using Rule-Based TODS.
"""
from botiverse.bots.basic_TODS.basic_TODS import BasicTODS

NAME = "boti"
domains_slots = {"book-flight": ["source", "destination", "time"]}
templates = {
                "book-flight":
                {
                    "source": ["Where do you want to fly from?",
                               "From where will you take the flight?"],
                    "destination": ["What is your destination?",
                                    "Where do you want to go?"],
                    "time": ["What time do you want to leave?"]
                }
            }
domains_pattern = {"book-flight": r"(i|I) want to (book|reserve) a? flights?"}
slots_pattern = {
                    "book-flight":
                    {
                        "source": r"from(?: city)? (cairo|giza)",
                        "destination": r"to(?: city)? (cairo|giza)",
                        "time": r"(saturday|sunday|monday|tuesday|wednesday|thursday|friday)"
                    }
                }

chatbot = BasicTODS(NAME, domains_slots, templates, domains_pattern, slots_pattern)

print(chatbot.get_dialogue_state())
print(chatbot.infer("I want to book a flight"))
print(chatbot.infer("I want to go to giza from cairo"))
print(chatbot.infer("on monday"))
print(chatbot.get_dialogue_state())
print(chatbot.reset())
print(chatbot.get_dialogue_state())



import numpy as np
import json
from gtts import gTTS
import tempfile
import os
from botiverse.models import TTS
from playsound import playsound

from botiverse.models import LSTMClassifier
from botiverse.preprocessors import Vocalize, Wav2Vec, Wav2Text, BertEmbedder, Frequency, BertSentenceEmbedder
from botiverse.bots.VoiceBot.utils import voice_input


class VoiceBot():
    '''An interface for the vocalizer chatbot which simulates a call with a customer service bot.'''
    def __init__(self,  call_json_path, repr='BERT-Sentence'):
        ''' 
        Load the call data from a json file.
        :param call_json_path: The path to the json file containing the call state machine.
        '''
        with open(call_json_path, 'r') as file:
            call_json = file.read()
        self.call_data = json.loads(call_json)
        self.current_node = 'A'
        self.wav2text = Wav2Text()
        if repr == 'BERT':
            self.bert_embeddings = BertEmbedder()
        elif repr == 'BERT-Sentence':
            self.bert_sentence_embeddings = BertSentenceEmbedder()
        else:
            raise Exception(f"Invalid representation {repr}. Expected BERT or BERT-Sentence.")        
    
    def generate_speech(self, text, offline=False):
        '''Use Google's TTS or offline FastSpeech 1.0 to play speech from the given text.
        :param text: The text to be converted into speech.
        :param offline: Whether to use offline FastSpeech 1.0 to generate speech.
        '''
        if offline:
            tts = TTS()
            tts.speak(text)
        else:
            tts = gTTS(text=text, lang='en', tld="us", slow=False)
            with tempfile.NamedTemporaryFile(suffix=".mp3", delete=False) as temp_audio:
                temp_filename = temp_audio.name
                tts.save(temp_filename)
                playsound(temp_filename)

    def simulate_call(self):
        '''
        Simulate a call with a customer service bot as driven by the call state machine.
        '''
        while True:
            if self.current_node == 'Z':
                # the final state has a different structure, bot only speaks and then the call ends
                bot_message = self.call_data[self.current_node]['Bot']
                self.generate_speech(bot_message)
                break

            # 1 - get the current node's data and from that get the message the bot should speak
            node_data = self.call_data[self.current_node]
            bot_message = node_data['Bot']
            self.generate_speech(bot_message)

            # 2 - get the intent options that the bot expects from the user and classify the user's response
            options = node_data['Options']
            intents = [option['Intent'] for option in options]
            max_dur = node_data['max_duration']
            human_resp = voice_input(record_time=int(max_dur))
            human_resp = self.wav2text.transcribe(human_resp)
            selected_ind, score = self.bert_sentence_embeddings.closest_sentence(human_resp, intents, retun_ind=True)
            print(f"you said: {human_resp} and the bot decided that you meant {intents[selected_ind]} with a score of {score}")
            
            # 3 - speak according to the chosen option
            speak_message = options[selected_ind]['Speak']
            self.generate_speech(speak_message)

            # 4 - go to the next state
            self.current_node = options[selected_ind]['Next']




from botiverse.bots.BasicBot.BasicBot import BasicBot
from botiverse.bots.WhizBot.WhizBot import WhizBot
from botiverse.bots.ConverseBot.ConverseBot import ConverseBot
from botiverse.bots.basic_TODS.basic_TODS import BasicTODS
from botiverse.bots.deep_TODS.deep_TODS import DeepTODS
from botiverse.bots.VoiceBot.SpeechClassifier import SpeechClassifier
from botiverse.bots.VoiceBot.VoiceBot import VoiceBot
